# Copyright (c) [2024] []
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Docstring in magnetism.
"""

import copy
import warnings
from os import environ
from typing import Union

import numpy as np
import sisl
from mpi4py import MPI
from scipy.special import roots_legendre
from tqdm import tqdm as tqdm

import grogupy.constants as constants
from grogupy.core import calc_Vu, onsite_projection
from grogupy.utilities import RotMa2b, spin_tracer, tau_u


class Hamiltonian:
    def __init__(self) -> None:
        self.H = None
        self.S = None
        self.NO = None
        self.cell = None
        self.nsc = None
        self.sc_off = None
        self.unit_cell_index = None
        self.G = None
        self.orientation = None
        self.hTRS = None
        self.hTRB = None
        self.XCF = None
        self.HK = None
        self.SK = None

    def copy(self):
        return copy.deepcopy(self)

    def rot_H_and_XCF(self, scf_xcf, orientation):
        # obtain rotated exchange field and Hamiltonian
        R: np.array = RotMa2b(scf_xcf, orientation)
        rot_XCF: np.array = np.einsum("ij,jklm->iklm", R, self.XCF)
        rot_H_XCF: np.array = sum(
            [
                np.kron(rot_XCF[i], tau)
                for i, tau in enumerate(
                    [constants.TAU_X, constants.TAU_Y, constants.TAU_Z]
                )
            ]
        )

        rot_H_XCF_uc: np.array = rot_H_XCF[self.unit_cell_index]

        # obtain total Hamiltonian with the rotated exchange field
        rot_H: np.array = self.hTRS + rot_H_XCF  # equation 76

        return rot_H, rot_H_XCF_uc

    def build_custom_H_and_S(
        self, dh: sisl.physics.Hamiltonian
    ) -> tuple[np.array, np.array]:
        """It builds the Hamiltonian and Overlap matrix from the sisl.dh class.

        It restructures the data in the SPIN BOX representation, where NS is
        the number of supercells and NO is the number of orbitals.

        Args:
            dh : sisl.physics.Hamiltonian
                Hamiltonian read in by sisl

        Returns:
            hh : (NS, NO, NO) np.array_like
                Hamiltonian in SPIN BOX representation
            ss : (NS, NO, NO) np.array_like
                Overlap matrix in SPIN BOX representation
        """

        self.NO = dh.no
        self.cell = dh.cell
        self.nsc = dh.lattice.nsc
        self.sc_off = dh.sc_off
        self.unit_cell_index = dh.lattice.sc_index([0, 0, 0])
        self.sislHamiltonian = dh

        NO = dh.no  # shorthand for number of orbitals in the unit cell

        # preprocessing Hamiltonian and overlap matrix elements
        h11 = dh.tocsr(dh.M11r)
        h11 += dh.tocsr(dh.M11i) * 1.0j
        h11 = (
            h11.toarray()
            .reshape(NO, dh.n_s, NO)
            .transpose(0, 2, 1)
            .astype("complex128")
        )

        h22 = dh.tocsr(dh.M22r)
        h22 += dh.tocsr(dh.M22i) * 1.0j
        h22 = (
            h22.toarray()
            .reshape(NO, dh.n_s, NO)
            .transpose(0, 2, 1)
            .astype("complex128")
        )

        h12 = dh.tocsr(dh.M12r)
        h12 += dh.tocsr(dh.M12i) * 1.0j
        h12 = (
            h12.toarray()
            .reshape(NO, dh.n_s, NO)
            .transpose(0, 2, 1)
            .astype("complex128")
        )

        h21 = dh.tocsr(dh.M21r)
        h21 += dh.tocsr(dh.M21i) * 1.0j
        h21 = (
            h21.toarray()
            .reshape(NO, dh.n_s, NO)
            .transpose(0, 2, 1)
            .astype("complex128")
        )

        sov = (
            dh.tocsr(dh.S_idx)
            .toarray()
            .reshape(NO, dh.n_s, NO)
            .transpose(0, 2, 1)
            .astype("complex128")
        )

        # Reorganization of Hamiltonian and overlap matrix elements to SPIN BOX representation
        U = np.vstack(
            [
                np.kron(np.eye(NO, dtype=int), [1, 0]),
                np.kron(np.eye(NO, dtype=int), [0, 1]),
            ]
        )
        # This is the permutation that transforms ud1ud2 to u12d12
        # That is this transforms FROM SPIN BOX to ORBITAL BOX => U
        # the inverse transformation is U.T u12d12 to ud1ud2
        # That is FROM ORBITAL BOX to SPIN BOX => U.T

        # From now on everything is in SPIN BOX!!
        hh, ss = np.array(
            [
                U.T
                @ np.block([[h11[:, :, i], h12[:, :, i]], [h21[:, :, i], h22[:, :, i]]])
                @ U
                for i in range(dh.lattice.nsc.prod())
            ]
        ), np.array(
            [
                U.T
                @ np.block(
                    [[sov[:, :, i], sov[:, :, i] * 0], [sov[:, :, i] * 0, sov[:, :, i]]]
                )
                @ U
                for i in range(dh.lattice.nsc.prod())
            ]
        )

        # copy arrays for tests
        hh_test, ss_test = hh.copy(), ss.copy()

        # symmetrizing Hamiltonian and Overlap matrix to make them hermitian
        for i in range(dh.lattice.sc_off.shape[0]):
            j = dh.lattice.sc_index(-dh.lattice.sc_off[i])
            h1, h1d = hh[i], hh[j]
            hh[i], hh[j] = (h1 + h1d.T.conj()) / 2, (h1d + h1.T.conj()) / 2
            s1, s1d = ss[i], ss[j]
            ss[i], ss[j] = (s1 + s1d.T.conj()) / 2, (s1d + s1.T.conj()) / 2

        # test symmetrization
        diff_hh = abs(hh - hh_test).max()
        diff_ss = abs(ss - ss_test).max()
        if diff_hh > 1e-12:
            warnings.warn(
                f"Hamiltonian changed after symmetrization. Largest change is {diff_hh}"
            )
        if diff_ss > 1e-12:
            warnings.warn(
                f"Overlap matrix changed after symmetrization. Largest change is {diff_hh}"
            )

        self.H = hh
        self.S = ss
        self.hTRS, self.hTRB, self.XCF = self._extract_exchange_field()

    def _extract_exchange_field(self) -> np.array:
        """Extracts the exchange field from the Hamiltonian.

        Args:
            hh : (NO, NO) np.array_like
                Hamiltonian in spin box representation
        Returns:
            np.array_like:
                XCF field
        """

        # identifying TRS and TRB parts of the Hamiltonian
        TAUY: np.array = np.kron(np.eye(self.NO), constants.TAU_Y)
        hTR: np.array = np.array(
            [TAUY @ self.H[i].conj() @ TAUY for i in range(self.nsc.prod())]
        )
        hTRS: np.array = (self.H + hTR) / 2
        hTRB: np.array = (self.H - hTR) / 2

        # extracting the exchange field
        traced: np.array = [
            spin_tracer(hTRB[i]) for i in range(self.nsc.prod())
        ]  # equation 77
        XCF: np.array = np.array(
            [
                np.array([f["x"] / 2 for f in traced]),
                np.array([f["y"] / 2 for f in traced]),
                np.array([f["z"] / 2 for f in traced]),
            ]
        )

        # check if exchange field has scalar part
        max_xcfs: float = abs(np.array(np.array([f["c"] / 2 for f in traced]))).max()
        if max_xcfs > 1e-12:
            warnings.warn(
                f"Exchange field has non negligible scalar part. Largest value is {max_xcfs}"
            )

        return hTRS, hTRB, XCF

    def _parallel_Gk(self, eran: np.array, eset: int) -> np.array:
        """Calculates the Greens function by inversion.

        It calculates the Greens function on all the energy levels at the same time.

        Args:
            HK : (NO, NO), np.array_like
                Hamiltonian at a given k point
            SK : (NO, NO), np.array_like
                Overlap Matrix at a given k point
            eran : (eset) np.array_like
                Energy sample along the contour
            eset : int
                Number of energy samples along the contour

        Returns:
            Gk : (eset, NO, NO), np.array_like
                Green's function at a given k point
        """

        # Calculates the Greens function on all the energy levels
        return np.linalg.inv(self.SK * eran.reshape(eset, 1, 1) - self.HK)

    def _sequential_Gk(self, eran: np.array, eset: int) -> np.array:
        """Calculates the Greens function by inversion.

        It calculates sequentially over the energy levels.

        Args:
            HK : (NO, NO), np.array_like
                Hamiltonian at a given k point
            SK : (NO, NO), np.array_like
                Overlap Matrix at a given k point
            eran : (eset) np.array_like
                Energy sample along the contour
            eset : int
                Number of energy samples along the contour

        Returns:
            Gk : (eset, NO, NO), np.array_like
                Green's function at a given k point
        """

        # creates an empty holder
        Gk = np.zeros((eset, self.HK.shape[0], self.HK.shape[1]), dtype="complex128")
        # fills the holder sequentially by the Greens function on a given energy
        for j in range(eset):
            Gk[j] = np.linalg.inv(self.SK * eran[j] - self.HK)

        return Gk

    def setup_at_k_point(self, k: tuple = (0, 0, 0)) -> tuple[np.array, np.array]:
        """Speed up Hk and Sk generation.

        Calculates the Hamiltonian and the Overlap matrix at a given k point. It is faster that the sisl version.

        Args:
            H : np.array_like
                Hamiltonian in spin box form
            ss : np.array_like
                Overlap matrix in spin box form
            sc_off : list
                supercell indices of the Hamiltonian
            k : tuple, optional
                The k point where the matrices are set up. Defaults to (0, 0, 0)

        Returns:
            np.array_like
                Hamiltonian at the given k point
            np.array_like
                Overlap matrix at the given k point
        """

        # this two conversion lines are from the sisl source
        k = np.asarray(k, np.float64)
        k.shape = (-1,)

        # this generates the list of phases
        phases = np.exp(-1j * 2 * np.pi * k @ self.sc_off.T)

        # phases applied to the hamiltonian
        self.HK = np.einsum("abc,a->bc", self.H, phases)
        self.SK = np.einsum("abc,a->bc", self.S, phases)


class Contour:
    def __init__(self, emin: float, eset: int, esetp: float, emax: float = 0) -> None:
        self.emin = emin
        self.emax = emax
        self.eset = eset
        self.esetp = esetp
        samples, weights = self._make_contour(emin, emax, eset, esetp)
        self.samples = samples
        self.weights = weights

    def _make_contour(
        self, emin: float, emax: float, enum: int, p: float
    ) -> tuple[np.array, float]:
        """A more sophisticated contour generator.

        Calculates the parameters for the complex contour integral. It uses the
        Legendre-Gauss quadrature method.   It returns a class that contains
        the information for the contour integral.

        Args:
            emin : int, optional
                Energy minimum of the contour. Defaults to -20
            emax : float, optional
                Energy maximum of the contour. Defaults to 0.0, so the Fermi level
            enum : int, optional
                Number of sample points along the contour. Defaults to 42
            p : int, optional
                Shape parameter that describes the distribution of the sample points. Defaults to 150

        Returns:
            Container
                Contains all the information for the contour integral
        """
        x, wl = roots_legendre(enum)
        R = (emax - emin) / 2  # radius
        z0 = (emax + emin) / 2  # center point
        y1 = -np.log(1 + np.pi * p)  # lower bound
        y2 = 0  # upper bound

        y = (y2 - y1) / 2 * x + (y2 + y1) / 2
        phi = (np.exp(-y) - 1) / p  # angle parameter
        ze = z0 + R * np.exp(1j * phi)  # complex points for path
        we = -(y2 - y1) / 2 * np.exp(-y) / p * 1j * (ze - z0) * wl

        return ze, we


class KSpace:
    def __init__(self, kdirs: str, kset: int, parallel_size: int = 1) -> None:
        self.kdirs = kdirs
        self.kset = kset
        self.kpoints = self._generate_k_points(self.kdirs, self.kset)
        self.weights = np.ones(len(self.kpoints)) / len(self.kpoints)
        self.__split_kpoints(parallel_size)

    def _generate_k_points(self, kdirs: str, kset: int) -> np.array:
        """Simple k-grid generator to sample the Brillouin zone.

        Depending on the value of the dirs
        argument k sampling in 1,2 or 3 dimensions is generated.
        If dirs argument does not contain either of x, y or z
        a kset of a single k-pont at the origin is returned. The total number of k points is the NUMK**(dimensions)


        Args:
            kdirs : str
                Directions of the k points in the Brillouin zone. They are the three lattice vectors
            kset : int
                The number of k points in a direction

        Returns:
            np.array_like
                An array of k points that uniformly sample the Brillouin zone in the given directions
        """

        # if there is no xyz in dirs return the Gamma point
        if not (sum([d in kdirs for d in "xyz"])):
            return np.array([[0, 0, 0]])

        kran = len(kdirs) * [np.linspace(0, 1, kset, endpoint=False)]
        mg = np.meshgrid(*kran)
        directions = dict()

        for d in enumerate(kdirs):
            directions[d[1]] = mg[d[0]].flatten()
        for d in "xyz":
            if not (d in kdirs):
                directions[d] = 0 * directions[kdirs[0]]
        kset = np.array([directions[d] for d in "xyz"]).T

        return kset

    def __split_kpoints(self, size):
        self.parallel_weights = np.array_split(self.weights, size)
        self.parallel_kpoints = np.array_split(self.kpoints, size)
        self.parallel_kpoints[0] = tqdm(self.parallel_kpoints[0])


class MagneticEntity:
    number_of_entities = 0

    def __init__(
        self,
        dh: sisl.physics.Hamiltonian,
        atom: Union[None, int, list] = None,
        l: Union[None, int, list] = None,
        orb: Union[None, int, list] = None,
    ) -> None:
        self.orbital_indices = self.__parse_for_orbital_indices(dh, atom, l, orb)
        self.atom = self.__parse_for_atom_indices(dh, self.orbital_indices)
        self.spin_box_indices = self.__blow_up_orbital_index(self.orbital_indices)
        self.SBS = len(self.spin_box_indices)
        self.xyz = [dh.xyz[i] for i in self.atom]
        try:
            self.tag = [f"[{i}]{dh.atoms[i].tag}({self.l[i]})" for i in self.atom]
        except:
            self.tag = [f"[{i}]{dh.atoms[i].tag}(Unknown)" for i in self.atom]
        self.Vu1 = []
        self.Vu2 = []
        self.Gii = []
        self._Gii_tmp = []
        self._energies = []
        self.K = []
        self.K_consistency = None

        MagneticEntity.number_of_entities += 1

    def __parse_for_atom_indices(self, dh, orbital_indices):
        atoms = []
        for orb in orbital_indices:
            atoms.append(dh.o2a(orb))
        atoms = np.sort(np.unique(np.array(atoms)))
        return atoms

    def __blow_up_orbital_index(self, orb_indices: np.array) -> np.array:
        """Function to blow up orbital indices to make SPIN BOX indices.

        Args:
            orb_indices : np.array_like
                These are the indices in ORBITAL BOX

        Returns:
            orb_indices : np.array_like
                These are the indices in SPIN BOX
        """

        orb_indices = np.array([[2 * o, 2 * o + 1] for o in orb_indices]).flatten()

        return orb_indices

    def __parse_for_orbital_indices(
        self,
        dh: Hamiltonian,
        atom: Union[None, int, list] = None,
        l: Union[None, int, list] = None,
        orb: Union[None, int, list] = None,
    ) -> np.array:
        """Function to define orbital indices of a given magnetic entity.

        There are five possible input types:
        1. Cluster: a list of atoms
        2. Atom: must be only one atom
        3. AtomShell: one atom and a list of shells indexed in the atom
        4. AtomOrbital: one atom and a list of orbitals indexed in the atom
        5. Orbitals: a list of orbitals  indexed in the Hamiltonian
        Everything else raises an Exception.

        Args:
            dh : sisl.physics.Hamiltonian
                Hamiltonian from sisl
            atom : integer or list of integers, optional
                Defining atom (or atoms) in the unit cell forming the magnetic entity. Defaults to None
            l : integer or list of integers, optional
                Defining the angular momentum channel. Defaults to None
            orb : integer or list of integers, optional
                Defining the orbital index in the Hamiltonian or on the atom. Defaults to None

        Returns:
            list
                The orbital indices of the given magnetic entity
        """

        # the case for the Orbitals keyword
        if atom is None and l is None:
            orbital_indices = orb

        # the case for the Cluster keyword
        elif isinstance(atom, list) and l is None and orb is None:
            orbital_indices = []
            for a in atom:
                a_orb_idx = dh.geometry.a2o(a, all=True)
                orbital_indices.append(a_orb_idx)
            orbital_indices = np.array(orbital_indices).flatten()

        # the case for the Atom keyword
        elif isinstance(atom, int) and l is None and orb is None:
            orbital_indices = dh.geometry.a2o(atom, all=True)

        # the case for the Atomshell keyword
        elif isinstance(atom, int) and l is not None and orb is None:
            orbital_indices = dh.geometry.a2o(atom, all=True)
            # make sure l is a list for the following step
            if isinstance(l, int):
                l = [l]
            mask = [orbital.l in l for orbital in dh.geometry.atoms[atom].orbitals]
            orbital_indices = orbital_indices[mask]

        # the case for the AtomOrbital keyword
        elif isinstance(atom, int) and l is None and orb is not None:
            orbital_indices = dh.geometry.a2o(atom, all=True)
            # make sure orb is a list for the following step
            if isinstance(orb, int):
                orb = [orb]
            orbital_indices = orbital_indices[orb]

        return orbital_indices  # numpy array containing integers labeling orbitals associated to a magnetic entity.

    def calculate_energies(self, weights):
        energies = []
        # iterate over the quantization axes
        for i, Gii in enumerate(self.Gii):
            storage: list = []
            # iterate over the first and second order local perturbations
            for Vu1, Vu2 in zip(self.Vu1[i], self.Vu2[i]):
                # The Szunyogh-Lichtenstein formula
                traced: np.array = np.trace(
                    (Vu2 @ Gii + 0.5 * Gii @ Vu1 @ Gii), axis1=1, axis2=2
                )  # this is the on site projection
                # evaluation of the contour integral
                storage.append(np.trapz(-1 / np.pi * np.imag(traced * weights)))
            # fill up the magnetic entities dictionary with the energies
            energies.append(storage)
        # convert to np array
        self._energies = np.array(energies)

    def calculate_anisotropy(self):
        """Calculates the renormalized anisotropy tensor from the energies.

        Args:
            mag_ent : dict
                An element from the magnetic entities

        Returns:
            K : np.array_like
                elements of the anisotropy tensor
            consistency_check : float
                Absolute value of the difference from the consistency check
        """

        # get the energies
        energies = self._energies

        K = np.zeros((3, 3))

        # calculate the diagonal tensor elements
        K[0, 0] = energies[1, 1] - energies[1, 0]
        K[1, 1] = energies[0, 1] - energies[0, 0]
        K[2, 2] = 0

        # calculate the off-diagonal tensor elements
        K[0, 1] = (energies[2, 0] + energies[2, 1]) / 2 - energies[2, 2]
        K[1, 0] = K[0, 1]
        K[0, 2] = (energies[1, 0] + energies[1, 1]) / 2 - energies[1, 2]
        K[2, 0] = K[0, 2]
        K[1, 2] = (energies[0, 0] + energies[0, 1]) / 2 - energies[0, 2]
        K[2, 1] = K[1, 2]

        # perform consistency check
        calculated_diff = K[1, 1] - K[0, 0]
        expected_diff = energies[2, 0] - energies[2, 1]
        consistency_check = abs(calculated_diff - expected_diff)

        self.K = K * sisl.unit_convert("eV", "meV")
        self.K_consistency = consistency_check * sisl.unit_convert("eV", "meV")

    def dictionary(self):
        out = dict()
        warnings.warn("Tag creation by orb index")

        out["orbital_indices"] = self.orbital_indices
        out["spin_box_indices"] = self.spin_box_indices
        out["xyz"] = self.xyz
        out["tag"] = self.tag
        out["energies"] = self.energies()

        return out


class Pair:
    number_of_pairs = 0

    def __init__(
        self,
        M1: MagneticEntity,
        M2: MagneticEntity,
        hamiltonian: Hamiltonian,
        supercell_shift: Union[list, np.array] = np.array([0, 0, 0]),
    ) -> None:
        self.M1 = M1
        self.M2 = M2
        self.supercell_shift = np.array(supercell_shift)
        self.supercell_shift_xyz = self.supercell_shift @ hamiltonian.cell
        self.xyz = np.array([M1.xyz, M2.xyz + self.supercell_shift_xyz])
        self.distance = np.linalg.norm(self.xyz[0] - self.xyz[1])
        self.SBS1 = M1.SBS
        self.SBS2 = M2.SBS
        self.tags = np.array([M1.tag, M2.tag])
        self.Gij = []
        self.Gji = []
        self._Gij_tmp = []
        self._Gji_tmp = []
        self._energies = []

        Pair.number_of_pairs += 1

    def calculate_energies(self, weights):
        energies = []
        # iterate over the quantization axes
        for i, (Gij, Gji) in enumerate(zip(self.Gij, self.Gji)):
            storage: list = []
            # iterate over the first order local perturbations in all possible orientations for the two sites
            # actually all possible orientations without the orientation for the off-diagonal anisotropy
            # that is why we only take the first two of each Vu1
            for Vui in self.M1.Vu1[i][:2]:
                for Vuj in self.M1.Vu2[i][:2]:
                    # The Szunyogh-Lichtenstein formula
                    traced: np.array = np.trace(
                        (Vui @ Gij @ Vuj @ Gji), axis1=1, axis2=2
                    )  # this is the on site projection
                    # evaluation of the contour integral
                    storage.append(np.trapz(-1 / np.pi * np.imag(traced * weights)))
            # fill up the pairs dictionary with the energies
            energies.append(storage)
        # convert to np array
        self._energies = np.array(energies)

    def calculate_exchange_tensor(self) -> tuple[float, np.array, np.array, np.array]:
        """Calculates the exchange tensor from the energies.

        It produces the isotropic exchange, the relevant elements
        from the Dzyaloshinskii-Morilla (Dm) tensor, the symmetric-anisotropy
        and the complete exchange tensor.

        Args:
            pair : dict
                An element from the pairs

        Returns:
            J_iso : float
                Isotropic exchange (Tr[J] / 3)
            J_S : np.array_like
                Symmetric-anisotropy (J_S = J - J_iso * I ––> Jxx, Jyy, Jxy, Jxz, Jyz)
            D : np.array_like
                DM elements (Dx, Dy, Dz)
            J : np.array_like
                Complete exchange tensor flattened (Jxx, Jxy, Jxz, Jyx, Jyy, Jyz, Jzx, Jzy, Jzz)
        """

        # energies from rotations
        energies = self._energies
        # Initialize output arrays
        J = np.zeros((3, 3))
        D = np.zeros(3)

        # J matrix calculations
        # J(1,1) = mean([DEij(2,2,2), DEij(2,2,3)])
        J[0, 0] = np.mean([energies[1, 3], energies[2, 3]])

        # J(1,2) = -mean([DEij(1,2,3), DEij(2,1,3)])
        J[0, 1] = -np.mean([energies[2, 1], energies[2, 2]])
        J[1, 0] = J[0, 1]

        # J(1,3) = -mean([DEij(1,2,2), DEij(2,1,2)])
        J[0, 2] = -np.mean([energies[1, 1], energies[1, 2]])
        J[2, 0] = J[0, 2]

        # J(2,2) = mean([DEij(2,2,1), DEij(1,1,3)])
        J[1, 1] = np.mean([energies[0, 3], energies[2, 0]])

        # J(2,3) = -mean([DEij(1,2,1), DEij(2,1,1)])
        J[1, 2] = -np.mean([energies[0, 1], energies[0, 2]])
        J[2, 1] = J[1, 2]

        # J(3,3) = mean([DEij(1,1,1), DEij(1,1,2)])
        J[2, 2] = np.mean([energies[0, 0], energies[1, 0]])

        # D vector calculations
        # D(1) = mean([DEij(1,2,1), -DEij(2,1,1)])
        D[0] = np.mean([energies[0, 1], -energies[0, 2]])

        # D(2) = mean([DEij(2,1,2), -DEij(1,2,2)])
        D[1] = np.mean([energies[1, 2], -energies[1, 1]])

        # D(3) = mean([DEij(1,2,3), -DEij(2,1,3)])
        D[2] = np.mean([energies[2, 1], -energies[2, 2]])

        J_iso = np.trace(J) / 3
        # based on the grogu output pdf
        # traceless symmetric exchange matrix:
        # Jxx, Jyy, Jxy, Jxz, Jyz
        J_S = np.array([J[0, 0] - J_iso, J[1, 1] - J_iso, J[0, 1], J[0, 2], J[1, 2]])

        self.J = J * sisl.unit_convert("eV", "meV")
        self.J_S = J_S * sisl.unit_convert("eV", "meV")
        self.J_iso = J_iso * sisl.unit_convert("eV", "meV")
        self.D = D * sisl.unit_convert("eV", "meV")

    def dictionary(self):
        out = dict()
        out["supercell_shift"] = self.supercell_shift
        out["supercell_shift_xyz"] = self.supercell_shift_xyz
        out["xyz"] = self.xyz
        out["distance"] = self.distance
        out["tags"] = self.tags
        out["energies"] = self.energies()

        return out


class Simulation:
    root_node = 0

    def __init__(self, simulation_parameters: dict) -> None:
        self.INFILE = simulation_parameters["INFILE"]
        self.EIGFILE = simulation_parameters["INFILE"][:-3] + "EIG"
        self.OUTFILE = simulation_parameters["OUTFILE"]
        self.SCF_XCF_ORIENTATION = simulation_parameters["SCF_XCF_ORIENTATION"]
        self.REF_XCF_ORIENTATIONS = simulation_parameters["REF_XCF_ORIENTATIONS"]
        self.KSET = simulation_parameters["KSET"]
        self.KDIRS = simulation_parameters["KDIRS"]
        self.EBOT = simulation_parameters["EBOT"]
        self.ESET = simulation_parameters["ESET"]
        self.ESETP = simulation_parameters["ESETP"]
        self.PARALLEL_SOLVER_FOR_GK = simulation_parameters["PARALLEL_SOLVER_FOR_GK"]
        self.EMAX = 0
        self.AUTOMATIC_EBOT = False

        self.magnetic_entities = []
        self.pairs = []

        if not self.OUTFILE.endswith(".pickle"):
            self.OUTFILE += ".pickle"

        if self.EBOT is None:
            self.EBOT = self.read_siesta_emin(self.EIGFILE)
            self.AUTOMATIC_EBOT = True

        try:
            self.SLURM_ID = environ["SLURM_JOB_ID"]
        except:
            self.SLURM_ID = "Could not be determined."
            # add third direction for off-diagonal anisotropy
        for orientation in self.REF_XCF_ORIENTATIONS:
            o1: np.array = orientation["vw"][0]
            o2: np.array = orientation["vw"][1]
            orientation["vw"] = np.vstack((orientation["vw"], (o1 + o2) / np.sqrt(2)))

        # MPI parameters
        self.comm = MPI.COMM_WORLD
        self.parallel_size = self.comm.Get_size()
        self.rank = self.comm.Get_rank()

    def __print__(self) -> None:
        """It prints the parameters and the description of the job.

        Args:
            self :
                It contains the simulations parameters
        """

        print(
            "================================================================================================================================================================"
        )
        try:
            print(f"SLURM job ID: {environ['SLURM_JOB_ID']}")
        except:
            print("SLURM job ID not found.")
        print("Input file: ")
        print(self.INFILE)
        print("Output file: ")
        print(self.OUTFILE)
        print(
            "Number of nodes in the parallel cluster: ",
            self.PARALLEL_SIZE,
        )
        if self.PARALLEL_SOLVER_FOR_GK:
            print("solver used for Greens function calculation: parallel")
        else:
            print("solver used for Greens function calculation: sequential")
        print(
            "================================================================================================================================================================"
        )
        print("Cell [Ang]: ")
        print("NOT YET")
        print(
            "================================================================================================================================================================"
        )
        print("DFT axis: ")
        print(self.SCF_XCF_ORIENTATION)
        print("Quantization axis and perpendicular rotation directions:")
        for ref in self.REF_XCF_ORIENTATIONS:
            print(ref["o"], " --» ", ref["vw"])
        print(
            "================================================================================================================================================================"
        )
        print("Parameters for the contour integral:")
        print("Number of k points: ", self.KSET)
        print("k point directions: ", self.KDIRS)
        if self.AUTOMATIC_EBOT:
            print(
                "Ebot: ",
                self.EBOT,
                "        WARNING: This was automatically determined!",
            )
        else:
            print("Ebot: ", self.EBOT)
        print("Eset: ", self.ESET)
        print("Esetp: ", self.ESETP)
        print(
            "================================================================================================================================================================"
        )

    def add_KSpace(self, kspace: KSpace):
        self.KSpace = kspace

    def add_Hamiltonian(self, hamiltonian: Hamiltonian):
        self.Hamiltonian = hamiltonian
        self.sislHamiltonian = hamiltonian.sislHamiltonian

    def add_Contour(self, contour: Contour):
        self.Contour = contour

    def add_magnetic_entities(self, magnetic_entities: Union[MagneticEntity, list]):
        if isinstance(magnetic_entities, MagneticEntity):
            magnetic_entities = [magnetic_entities]

        self.magnetic_entities = magnetic_entities

    def add_pairs(self, pairs: Union[Pair, dict, list]):
        if isinstance(pairs, Pair):
            self.pairs.append(pairs)

        elif isinstance(pairs, dict):
            self.pairs.append(
                Pair(
                    self.magnetic_entities[pairs["ai"]],
                    self.magnetic_entities[pairs["aj"]],
                    self.sislHamiltonian,
                    pairs["Ruc"],
                )
            )

        elif isinstance(pairs, list):
            for pair in pairs:
                if isinstance(pair, Pair):
                    self.pairs.append(pair)
                elif isinstance(pair, dict):
                    self.pairs.append(
                        Pair(
                            self.magnetic_entities[pair["ai"]],
                            self.magnetic_entities[pair["aj"]],
                            self.sislHamiltonian,
                            pair["Ruc"],
                        )
                    )
                else:
                    raise Exception("Bad input")
        else:
            raise Exception("Bad input")

    def read_siesta_emin(self, eigfile: str) -> float:
        """It reads the lowest energy level from the siesta run.

        It uses the .EIG file from siesta that contains the eigenvalues.

        Args:
            eigfile : str
                The path to the .EIG file

        Returns:
            float
                The energy minimum
        """

        # read the file
        eigenvalues = sisl.io.eigSileSiesta(eigfile).read_data()

        return eigenvalues.min()

    def rotate_XCF(self):
        self.rotated_hamiltonians = []
        for mag_ent in self.magnetic_entities:
            for i in range(len(self.REF_XCF_ORIENTATIONS)):
                mag_ent.Vu1.append([])
                mag_ent.Vu2.append([])
                mag_ent.Gii.append(
                    np.zeros((self.ESET, mag_ent.SBS, mag_ent.SBS), dtype="complex128")
                )
                mag_ent._Gii_tmp.append(
                    np.zeros((self.ESET, mag_ent.SBS, mag_ent.SBS), dtype="complex128")
                )
        for pair in self.pairs:
            for i in range(len(self.REF_XCF_ORIENTATIONS)):
                pair.Gij.append(
                    np.zeros((self.ESET, pair.SBS1, pair.SBS2), dtype="complex128")
                )
                pair._Gij_tmp.append(
                    np.zeros((self.ESET, pair.SBS1, pair.SBS2), dtype="complex128")
                )
                pair.Gji.append(
                    np.zeros((self.ESET, pair.SBS2, pair.SBS1), dtype="complex128")
                )
                pair._Gji_tmp.append(
                    np.zeros((self.ESET, pair.SBS2, pair.SBS1), dtype="complex128")
                )
        # iterate over the reference directions (quantization axes)
        for i, orient in enumerate(self.REF_XCF_ORIENTATIONS):
            rot_H, rot_H_XCF_uc = self.Hamiltonian.rot_H_and_XCF(
                self.SCF_XCF_ORIENTATION, orient["o"]
            )

            # store the relevant information of the Hamiltonian
            current_hamiltonian = Hamiltonian()
            current_hamiltonian.H = rot_H
            current_hamiltonian.S = self.Hamiltonian.S
            current_hamiltonian.sc_off = self.Hamiltonian.sc_off

            self.rotated_hamiltonians.append(current_hamiltonian)

            # these are the rotations (for now) perpendicular to the quantization axis
            for u in orient["vw"]:
                # section 2.H
                Tu: np.array = np.kron(np.eye(self.Hamiltonian.NO, dtype=int), tau_u(u))
                Vu1, Vu2 = calc_Vu(rot_H_XCF_uc, Tu)

                for mag_ent in self.magnetic_entities:
                    # fill up the perturbed potentials (for now) based on the on-site projections
                    mag_ent.Vu1[i].append(
                        onsite_projection(
                            Vu1, mag_ent.spin_box_indices, mag_ent.spin_box_indices
                        )
                    )
                    mag_ent.Vu2[i].append(
                        onsite_projection(
                            Vu2, mag_ent.spin_box_indices, mag_ent.spin_box_indices
                        )
                    )

    def solve(self):
        # sampling the integrand on the contour and the BZ
        for i, k in enumerate(self.KSpace.parallel_kpoints[self.rank]):
            # weight of k point in BZ integral
            wk: float = self.KSpace.parallel_weights[self.rank][i]
            # iterate over reference directions
            for i, hamiltonian_orientation in enumerate(self.rotated_hamiltonians):
                # calculate Hamiltonian and Overlap matrix in a given k point
                hamiltonian_orientation.setup_at_k_point(k)

                if self.PARALLEL_SOLVER_FOR_GK:
                    Gk = hamiltonian_orientation._parallel_Gk(
                        self.Contour.samples, self.Contour.eset
                    )
                else:
                    # solve Greens function sequentially for the energies, because of memory bound
                    Gk = hamiltonian_orientation._sequential_Gk(
                        self.Contour.samples, self.Contour.eset
                    )

                # store the Greens function slice of the magnetic entities
                for mag_ent in self.magnetic_entities:
                    mag_ent._Gii_tmp[i] += (
                        onsite_projection(
                            Gk, mag_ent.spin_box_indices, mag_ent.spin_box_indices
                        )
                        * wk
                    )

                for pair in self.pairs:
                    # add phase shift based on the cell difference
                    phase: np.array = np.exp(
                        1j * 2 * np.pi * k @ pair.supercell_shift.T
                    )

                    # store the Greens function slice of the magnetic entities
                    pair._Gij_tmp[i] += (
                        onsite_projection(
                            Gk, pair.M1.spin_box_indices, pair.M2.spin_box_indices
                        )
                        * phase
                        * wk
                    )
                    pair._Gji_tmp[i] += (
                        onsite_projection(
                            Gk, pair.M2.spin_box_indices, pair.M1.spin_box_indices
                        )
                        / phase
                        * wk
                    )

        # sum reduce partial results of mpi nodes
        for i in range(len(self.rotated_hamiltonians)):
            for mag_ent in self.magnetic_entities:
                self.comm.Reduce(
                    mag_ent._Gii_tmp[i], mag_ent.Gii[i], root=self.root_node
                )

            for pair in self.pairs:
                self.comm.Reduce(pair._Gij_tmp[i], pair.Gij[i], root=self.root_node)
                self.comm.Reduce(pair._Gji_tmp[i], pair.Gji[i], root=self.root_node)

        for mag_ent in self.magnetic_entities:
            mag_ent.calculate_energies(self.Contour.weights)
            mag_ent.calculate_anisotropy()
        for pair in self.pairs:
            pair.calculate_energies(self.Contour.weights)
            pair.calculate_exchange_tensor()
