# Copyright (c) [2024] []
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Docstring in utilities.
"""

from typing import Union

import numpy as np

from grogupy.constants import TAU_0, TAU_X, TAU_Y, TAU_Z


def spin_tracer(M: np.array) -> dict:
    """Spin tracer utility.

    This takes an operator with the orbital-spin sequence:
    orbital 1 up,
    orbital 1 down,
    orbital 2 up,
    orbital 2 down,
    that is in the SPIN-BOX representation,
    and extracts orbital dependent Pauli traces.


    Args:
        M : np.array_like
            Traceable matrix

    Returns:
        dict
            It contains the traced matrix with "x", "y", "z" and "c"
    """
    M11 = M[0::2, 0::2]
    M12 = M[0::2, 1::2]
    M21 = M[1::2, 0::2]
    M22 = M[1::2, 1::2]

    M_o = dict()
    M_o["x"] = M12 + M21
    M_o["y"] = 1j * (M12 - M21)
    M_o["z"] = M11 - M22
    M_o["c"] = M11 + M22

    return M_o


def commutator(a: np.array, b: np.array) -> np.array:
    """Shorthand for commutator.

    Commutator of two matrices in the mathematical sense.

    Args:
        a : np.array_like
            The first matrix
        b : np.array_like
            The second matrix

    Returns:
        np.array_like
            The commutator of a and b
    """

    return a @ b - b @ a


# define some useful functions


def tau_u(u: Union[list, np.array]) -> np.array:
    """Pauli matrix in direction u.

    Returns the vector u in the basis of the Pauli matrices.

    Args:
        u : list or np.array_like
            The direction

    Returns:
        np.array_like
            Arbitrary direction in the base of the Pauli matrices
    """

    # u is force to be of unit length
    u = u / np.linalg.norm(u)

    return u[0] * TAU_X + u[1] * TAU_Y + u[2] * TAU_Z


#
def crossM(u: Union[list, np.array]) -> np.array:
    """Definition for the cross-product matrix.

    It acts as a cross product with vector u.

    Args:
        u : list or np.array_like
            The second vector in the cross product

    Returns:
        np.array_like
            The matrix that represents teh cross product with a vector
    """
    return np.array([[0, -u[2], u[1]], [u[2], 0, -u[0]], [-u[1], u[0], 0]])


def RotM(theta: float, u: np.array, eps: float = 1e-10) -> np.array:
    """Definition of rotation matrix with angle theta around direction u.


    Args:
        theta : float
            The angle of rotation
        u : np.array_like
            The rotation axis
        eps : float, optional
            Cutoff for small elements in the resulting matrix. Defaults to 1e-10

    Returns:
        np.array_like
            The rotation matrix
    """

    u = u / np.linalg.norm(u)

    M = (
        np.cos(theta) * np.eye(3)
        + np.sin(theta) * crossM(u)
        + (1 - np.cos(theta)) * np.outer(u, u)
    )

    # kill off small numbers
    M[abs(M) < eps] = 0.0
    return M


def RotMa2b(a: np.array, b: np.array, eps: float = 1e-10) -> np.array:
    """Definition of rotation matrix rotating unit vector a to unit vector b.

    Function returns array R such that R @ a = b holds.

    Args:
        a : np.array_like
            First vector
        b : np.array_like
            Second vector
        eps : float, optional
            Cutoff for small elements in the resulting matrix. Defaults to 1e-10

    Returns:
        np.array_like
            The rotation matrix with the above property
    """

    v = np.cross(a, b)
    c = a @ b
    M = np.eye(3) + crossM(v) + crossM(v) @ crossM(v) / (1 + c)

    # kill off small numbers
    M[abs(M) < eps] = 0.0
    return M
